// Copyright (c) 1995, 1996 Regents of the University of California.
// All rights reserved.
//
// This software was developed by the Arcadia project
// at the University of California, Irvine.
//
// Redistribution and use in source and binary forms are permitted
// provided that the above copyright notice and this paragraph are
// duplicated in all such forms and that any documentation,
// advertising materials, and other materials related to such
// distribution and use acknowledge that the software was developed
// by the University of California, Irvine.  The name of the
// University may not be used to endorse or promote products derived
// from this software without specific prior written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
// IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
// WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

// File: FigPoly.java
// Classes: FigPoly
// Original Author: jrobbins@ics.uci.edu
// $Id: FigPoly.java,v 1.2 1997/01/24 19:15:53 jrobbins Exp $

package uci.graphedit;

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.util.*;

/** Primitive Fig to draw Polygons on a LayerDiagram */

public class FigPoly extends Fig {

  /** number of handles on the polygon, TOTALLY WRONG!!! */
  public static int numHandles = 8;

  /** The index of the last handle that the user touched. Pressing
   * delete while a handle is selected will remove one point from the
   * polygon. A value of -1 means that no handle is selected. */
  public int _selectedHandle = -1;

  /** The total number of points. */
  public int npoints = 0;

  /** The array of x coordinates. */
  public int xpoints[] = new int[4];

  /** The array of y coordinates. */
  public int ypoints[] = new int[4];


  /** Construct a new FigPoly w/ the given line color,
   * and fill color*/
  public FigPoly(Color line_color, Color fill_color) {
    super(0, 0, 50, 50, line_color, fill_color, numHandles);
  }

  /** Construct a new FigPoly w/ the given line color */
  public FigPoly(Color line_color) {
    this(line_color, null);
  }

  /** Construct a new FigPoly w/ the given attrobutes */
  public FigPoly(Hashtable gAttrs){
    this(null, null);
    setGraphicAttributes(gAttrs);
  }

  /** Construct a new FigPoly w/ the given point and attributes */
  public FigPoly(int x, int y, Hashtable gAttrs){
    this(null, null);
    setGraphicAttributes(gAttrs);
    addPoint(x, y);
  }

  ////////////////////////////////////////////////////////////////
  // Accessor functions

  /** Get the current vector of points. */
  public Polygon polygon() { return new Polygon(xpoints, ypoints, npoints); }

  /** Set the current vector of points. */
  public void polygon(Polygon p) {
    npoints = p.npoints;
    xpoints = new int[npoints];
    ypoints = new int[npoints];
    System.arraycopy(p.xpoints, 0, xpoints, 0, npoints);
    System.arraycopy(p.ypoints, 0, ypoints, 0, npoints);
    measure();
  }

  ////////////////////////////////////////////////////////////////
  // Geomertry related manipulations

  /** Change the position of the object from were it is
   * to were it is plus dx or dy. Often called when an object is
   * dragged. This could be very useful if local-coordinate systems are
   * used because deltas need less transforming... maybe.
   */
  void translate(int dx, int dy) {
    for (int i = 0; i < npoints; ++i) {
      xpoints[i] += dx;
      ypoints[i] += dy;
    }
    // there is no need to call measure
    Point p = position();
    p.x += dx; p.y += dy;
    position(p);
  }


  public void addPoint(int x, int y) {
    if (npoints == xpoints.length) {
      int tmp[];

      tmp = new int[npoints * 2];
      System.arraycopy(xpoints, 0, tmp, 0, npoints);
      xpoints = tmp;

      tmp = new int[npoints * 2];
      System.arraycopy(ypoints, 0, tmp, 0, npoints);
      ypoints = tmp;
    }
    xpoints[npoints] = x;
    ypoints[npoints] = y;
    _selectedHandle = npoints;
    npoints++;
    measure();
  }


  public final void addPoint(Point p) { addPoint(p.x, p.y); }

  public void moveVertex(int index, int x, int y) {
    xpoints[index] = x;
    ypoints[index] = y;
    measure();
  }

  public void removeSelectedVertex() {
    if (_selectedHandle == -1 || npoints < 3) return;
    int tmp[] = new int[npoints];
    int sh = _selectedHandle;
    System.arraycopy(xpoints, sh+1, tmp, 0, npoints - sh - 1);
    System.arraycopy(tmp, 0, xpoints, sh, npoints - sh - 1);
    System.arraycopy(ypoints, sh+1, tmp, 0, npoints - sh - 1);
    System.arraycopy(tmp, 0, ypoints, sh, npoints - sh - 1);
    --npoints;
    measure();
    _selectedHandle = _selectedHandle % npoints;
  }

  ////////////////////////////////////////////////////////////////
  // Event handling and related functions

  /** If teh user presses delete or backaspace while a handle is
   * selected, remove that point from the polygon. */
  public boolean keyDown(Event e, int key) {
    Editor ce = Globals.curEditor();
    if (_selectedHandle != -1 && (key == 127 || key == 8)) {
      ce.executeAction(new ActionRemoveVertex(),e);
      return true;
    }
    if (_selectedHandle != -1 && key == 9 && e.controlDown()) {
      // Needs-More-Work: the following should be in an Action.
      // ce.executeAction(new ActionSelectNextVertex(), e);
      startTrans();
      _selectedHandle = (_selectedHandle + 1) % npoints;
      endTrans();
      return true;
    }
    return super.keyDown(e, key);
  }

  /** When the user drags the handles of a polygon, move individual points */
  public void dragHandle(int mX, int mY, int anX, int anY, int handle) {
    _selectedHandle = handle;
    if (handle == -1) { super.dragHandle(mX, mY, anX, anY, handle); return; }
    moveVertex(handle, mX, mY);
  }



  ////////////////////////////////////////////////////////////////
  // Drawing methods

  public void draw(Graphics g) {
    if (objectFillColor != null) {
      g.setColor(objectFillColor);
      g.fillPolygon(xpoints, ypoints, npoints);
    }
    if (objectLineColor != null) {
      g.setColor(objectLineColor);
      g.drawPolygon(xpoints, ypoints, npoints);
    }
  }

  public void drawSelected(Graphics g) { drawHandles(g); }


  public void drawHandles(Graphics g) {
    g.setColor(Globals.prefs().handleColor());
    for (int i = 0; i < npoints; ++i)
      g.fillRect(xpoints[i] - HAND_SIZE/2, ypoints[i] - HAND_SIZE/2,
		 HAND_SIZE, HAND_SIZE);
    if (_selectedHandle != -1 )
      g.drawRect(xpoints[_selectedHandle] - HAND_SIZE/2 - 2,
		 ypoints[_selectedHandle] - HAND_SIZE/2 - 2,
		 HAND_SIZE + 3, HAND_SIZE + 3);
  }

  ////////////////////////////////////////////////////////////////
  // Methods related to selection
  public Selection selectionObject() {
    // no one handle is selected when the polygon is first selected
    _selectedHandle = -1;
    return new SelectionHandles(this);
  }

  /** Return a handle ID for the handle under the mouse, or -1 if
   * none. Needs-More-Work: in the future, return a Handle instance or
   * null. */
  public int pickHandle(int x, int y) {
    int xs[] = xpoints;
    int ys[] = ypoints;
    for (int i = 0; i < npoints; ++i) {
      if (x >= xs[i] - HAND_SIZE/2 && y >= ys[i] - HAND_SIZE/2 &&
	      x <= xs[i] + HAND_SIZE/2 && y <= ys[i] + HAND_SIZE/2) {
	    _selectedHandle = i;
	    return i;
	      }
	}
    _selectedHandle = -1;
    return -1;
  }

  public boolean inside(int x, int y) {
    Polygon p = polygon();
    return p.inside(x, y) || p.inside(x - GRIP_MARGIN, y) || p.inside(x + GRIP_MARGIN, y) ||
            p.inside(x , y - GRIP_MARGIN) || p.inside(x, y + GRIP_MARGIN);
  }

  ////////////////////////////////////////////////////////////////
  // Internal utility functions

  /** Deterime my position, width, and height from the points in the polygon. */
  protected void measure() {
    int minx = Integer.MAX_VALUE;
    int miny = Integer.MAX_VALUE;
    int maxx = Integer.MIN_VALUE;
    int maxy = Integer.MIN_VALUE;

    for (int i = 0; i < npoints; ++i) {
      int x = xpoints[i];
      minx = Math.min(minx, x);
      maxx = Math.max(maxx, x);
      int y = ypoints[i];
      miny = Math.min(miny, y);
      maxy = Math.max(maxy, y);
    }

    position(minx, miny);
    objectWidth = maxx - minx;
    objectHeight = maxy - miny;
  }

} /* end class FigPoly */
